package com.hncy.sjdz.util;

import cn.hutool.core.util.CoordinateUtil;
import org.geotools.geojson.geom.GeometryJSON;
import org.geotools.geometry.jts.JTS;
import org.locationtech.jts.geom.*;
import org.locationtech.jts.io.ParseException;
import org.locationtech.jts.io.WKTReader;
import org.locationtech.jts.io.WKTWriter;
import org.locationtech.jts.operation.overlay.snap.GeometrySnapper;
import org.locationtech.jts.simplify.DouglasPeuckerSimplifier;
import org.springframework.util.CollectionUtils;

import java.io.StringReader;
import java.math.BigDecimal;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 几何工具类
 *
 * @author
 * @date
 */
public class JTSGeometryUtils {

    private static GeometryFactory geometryFactory = new GeometryFactory(new PrecisionModel(PrecisionModel.FLOATING));

    private static WKTReader reader = new WKTReader(geometryFactory);

    private static WKTWriter writer = new WKTWriter();

    /**
     * 将前端坐标几何转为多边形字符串 POLYGON MULTIPOLYGON
     * @param webStr
     * @return
     */
    public static String webStrToWellKnownText(String webStr) {
        try {
            // 使用两个分号来分割不同的多边形
            String[] polygonStrings = webStr.split(";;");
            List<Geometry> geometries = new ArrayList<>();
            for (String p : polygonStrings) {
                // 遍历图形
                String[] coordinatePairs = p.split(";");
                Coordinate[] coordinates = new Coordinate[coordinatePairs.length];
                for (int i = 0; i < coordinatePairs.length; i++) {
                    String[] xy = coordinatePairs[i].split(",");
                    coordinates[i] = new Coordinate(Double.parseDouble(xy[0]), Double.parseDouble(xy[1]));
                }
                // 确保首尾坐标一致（闭合）
                LinearRing ring = geometryFactory.createLinearRing(coordinates);
                Polygon polygon = geometryFactory.createPolygon(ring);

                geometries.add(polygon);
            }

            Geometry result;
            if (geometries.size() > 1) {
                result = geometryFactory.createMultiPolygon(geometries.toArray(new Polygon[0]));
            }
            else {
                result = geometries.get(0);
            }
            // 简化的几何对象坐标
            DouglasPeuckerSimplifier simplifier = new DouglasPeuckerSimplifier(result);
            Geometry simplifiedGeometry = simplifier.getResultGeometry();
            return writer.write(simplifiedGeometry);
            // return writer.write(result);
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IllegalArgumentException("几何坐标异常！");
        }
    }

    /**
     * 将多边形字符串转为前端所需格式
     * @param wellKnownText
     * @return
     */
    public static String wellKnownTextToWebStr(String wellKnownText) {
        try {
            Geometry geometry = reader.read(wellKnownText);
            if (geometry instanceof Polygon) {
                Polygon polygon = (Polygon) geometry;
                // 转换为所需的字符串格式
                StringBuffer sb = new StringBuffer();
                polygonToWebStr(polygon, sb);
                String result = sb.toString();
                return result;
            }
            if (geometry instanceof GeometryCollection) {
                MultiPolygon multiPolygon = (MultiPolygon) geometry;
                StringBuffer sb = new StringBuffer();
                for (int k = 0; k < multiPolygon.getNumGeometries(); k++) {
                    Polygon polygon = (Polygon) multiPolygon.getGeometryN(k);
                    polygonToWebStr(polygon, sb);
                    // 在子多边形之间用两个分号分隔开子多边形
                    if (k < multiPolygon.getNumGeometries() - 1) {
                        sb.append(";;");
                    }
                }
                String result = sb.toString();
                return result;
            }
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IllegalArgumentException("几何坐标异常！！");
        }
        return null;
    }

    /**
     * 将polygon字符串转为前端所需格式
     * @param polygon
     * @param sb
     */
    private static void polygonToWebStr(Polygon polygon, StringBuffer sb) {
        // 遍历所有环（包括外环和内环）
        for (int i = 0; i < polygon.getNumInteriorRing() + 1; i++) {
            LinearRing ring = i == 0 ? polygon.getExteriorRing() : polygon.getInteriorRingN(i - 1);
            Coordinate[] coordinates = ring.getCoordinates();
            for (int j = 0; j < coordinates.length; j++) {
                Coordinate coordinate = coordinates[j];
                sb.append(coordinate.x).append(",").append(coordinate.y);
                // 如果不是最后一个坐标 添加分号
                if (j < coordinates.length - 1) {
                    sb.append(";");
                }
            }
            // 内环用|分割
            if (i >= 0 && i < polygon.getNumInteriorRing()) {
                sb.append("|");
            }
        }
    }

    /**
     * 将多边形字符串转为Polygon多边形
     * @param wellKnownText
     * @return
     */
    public static Polygon wellKnownTextToPolygon(String wellKnownText) {
        try {
            return (Polygon) reader.read(wellKnownText);
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IllegalArgumentException("多边形坐标异常！");
        }
    }

    /**
     * 将多边形字符串转为多边形
     * @param wellKnownText
     * @return
     */
    public static Geometry wellKnownTextToGeometry(String wellKnownText) {
        try {
            Geometry geometry = reader.read(wellKnownText);
            return geometry;
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IllegalArgumentException("多边形坐标异常！");
        }
    }

    /**
     * 将多边形字符串转为线
     * @param wellKnownText
     * @return
     */
    public static LineString wellKnownTextToLineString(String wellKnownText) {
        try {
            return (LineString) reader.read(wellKnownText);
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IllegalArgumentException("多边形坐标异常！");
        }
    }

    /**
     * 将Geometry对象转为GeoJson
     * @param geometry
     * @return
     */
    public static String geometryToGeoJson(Geometry geometry) {
        try {
            GeometryJSON geometryJSON = new GeometryJSON();
            return geometryJSON.toString(geometry);
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IllegalArgumentException("将Geometry转换为GeoJSON失败！");
        }
    }

    /**
     * 将GeoJson对象转为Geometry
     * @param geojson
     * @return
     */
    public static Geometry geojson2Geometry(String geojson) {
        try {
            GeometryJSON geometryJSON = new GeometryJSON();
            return geometryJSON.read(new StringReader(geojson));
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IllegalArgumentException("将GeoJSON转换为Geometry失败！");
        }
    }

    /**
     * 判断线是否与多边形至少有两个交点
     * @param line
     * @param geometry
     * @return
     */
    public static boolean hasAtLeastTwoIntersections(LineString line, Geometry geometry) {
        // 交点
        Set<Coordinate> uniqueIntersections = new HashSet<>();
        // 处理多边形或多面体情况
        if (geometry instanceof Polygon) {
            Polygon polygon = (Polygon) geometry;
            setUniqueIntersections(line, polygon, uniqueIntersections);
        }
        else if (geometry instanceof MultiPolygon) {
            MultiPolygon multiPolygon = (MultiPolygon) geometry;
            for (int i = 0; i < multiPolygon.getNumGeometries(); i++) {
                Polygon polygon = (Polygon) multiPolygon.getGeometryN(i);
                setUniqueIntersections(line, polygon, uniqueIntersections);
                // 当发现至少有两个交点时，可以直接返回true
                if (uniqueIntersections.size() >= 2) {
                    return true;
                }
            }
        }
        else {
            throw new IllegalArgumentException("几何图形格式异常 不属于Polygon或MultiPolygon");
        }
        return uniqueIntersections.size() >= 2;
    }

    /**
     * 设置交点
     * @param line
     * @param polygon
     * @param uniqueIntersections
     */
    private static void setUniqueIntersections(LineString line, Polygon polygon, Set<Coordinate> uniqueIntersections) {
        LinearRing exteriorRing = polygon.getExteriorRing();
        // 计算折线与外环的交点
        Coordinate[] intersectionPoints = line.intersection(exteriorRing).getCoordinates();
        for (Coordinate point : intersectionPoints) {
            uniqueIntersections.add(point);
        }

        // 处理内环（如果有）
        for (int i = 0; i < polygon.getNumInteriorRing(); i++) {
            LinearRing interiorRing = polygon.getInteriorRingN(i);
            // 计算折线与当前内环的交点
            intersectionPoints = line.intersection(interiorRing).getCoordinates();
            for (Coordinate point : intersectionPoints) {
                uniqueIntersections.add(point);
                // 当发现至少有两个交点时，可以提前结束循环
                if (uniqueIntersections.size() >= 2) {
                    return;
                }
            }
        }
    }

    /**
     * 判断线的起点和终点不在图形内
     * @param line 线
     * @param geometry 面
     * @return 如果起点和终点都不在几何图形内，则返回true；否则返回false
     */
    public static boolean isStartAndEndpointNotInGeometry(Geometry line, Geometry geometry) {
        Coordinate[] linePoints = line.getCoordinates();
        // 将起点和终点坐标转换为Point对象
        Point start = geometry.getFactory().createPoint(linePoints[0]);
        Point end = geometry.getFactory().createPoint(linePoints[linePoints.length - 1]);

        // 处理多边形或多面体情况
        if (geometry instanceof Polygon) {
            return !geometry.contains(start) && !geometry.contains(end);
        }
        else if (geometry instanceof MultiPolygon) {
            for (int i = 0; i < geometry.getNumGeometries(); i++) {
                Polygon polygon = (Polygon) geometry.getGeometryN(i);
                if (polygon.contains(start) || geometry.contains(end)) {
                    return false;
                }
            }
            return true;
        }
        else {
            throw new IllegalArgumentException("几何图形格式异常 不属于Polygon或MultiPolygon");
        }
    }

    /**
     * 根据coordinate坐标list创建Polygon
     * @param coordinates
     * @return
     */
    public static Polygon createPolygon(List<Coordinate> coordinates) {
        try {
            Coordinate[] coordinatesOneArr = coordinates.toArray(new Coordinate[coordinates.size()]);
            LinearRing shell = geometryFactory.createLinearRing(coordinatesOneArr);
            Polygon polygon = geometryFactory.createPolygon(shell);
            return polygon;
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IllegalArgumentException("创建Polygon失败！");
        }
    }

    public static Point toPoint(double longitude, double latitude) {
        // POINT(Longitude Latitude)
        return toPoint(new Coordinate(longitude, latitude));
    }

    public static Point toPoint(Coordinate coordinate) {
        // POINT(Longitude Latitude)
        return geometryFactory.createPoint(coordinate);
    }

    public static Double getDistance(Coordinate[] array) {
        try {
            LineString line = geometryFactory.createLineString(array);
            return line.getLength();
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IllegalArgumentException("获取距离失败！");
        }
    }

    /**
     * 根据坐标数组返回LineString
     * @param array
     * @return
     */
    public static LineString getLineString(Coordinate[] array) {
        try {
            LineString line = geometryFactory.createLineString(array);
            return line;
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IllegalArgumentException("创建LineString失败！");
        }
    }

    /**
     * 判断坐标点是否在多坐标点组成的多边形面内
     * @param longitude
     * @param latitude
     * @param wellKnownText
     * @return
     */
    public static boolean withinAndIntersects(double longitude, double latitude, String wellKnownText) {
        Polygon polygon = wellKnownTextToPolygon(wellKnownText);
        // 小于3个点无法组成多边形
        if (Objects.nonNull(polygon) && polygon.getCoordinates().length >= 3) {
            Point point = toPoint(longitude, latitude);
            if (point.within(polygon) || point.intersects(polygon)) {
                return true;
            }
        }
        return false;
    }

    /**
     * 切割多边形：计算封套的中线并将新的封套与多边形相交
     * @param polygon
     * @return
     */
    public static List<Geometry> split(Polygon polygon) {
        List<Geometry> ret = new ArrayList<>();
        final Envelope envelope = polygon.getEnvelopeInternal();
        double minX = envelope.getMinX();
        double maxX = envelope.getMaxX();
        double midX = minX + (maxX - minX) / 2.0;
        double minY = envelope.getMinY();
        double maxY = envelope.getMaxY();
        double midY = minY + (maxY - minY) / 2.0;
        Envelope llEnv = new Envelope(minX, midX, minY, midY);
        Envelope lrEnv = new Envelope(midX, maxX, minY, midY);
        Envelope ulEnv = new Envelope(minX, midX, midY, maxY);
        Envelope urEnv = new Envelope(midX, maxX, midY, maxY);
        Geometry ll = JTS.toGeometry(llEnv).intersection(polygon);
        Geometry lr = JTS.toGeometry(lrEnv).intersection(polygon);
        Geometry ul = JTS.toGeometry(ulEnv).intersection(polygon);
        Geometry ur = JTS.toGeometry(urEnv).intersection(polygon);
        ret.add(ll);
        ret.add(lr);
        ret.add(ul);
        ret.add(ur);
        return ret;
    }

    public static List<Geometry> split(Polygon polygon, int meshSize) {
        List<Geometry> ret = new ArrayList<>();
        // meshSize 并不能保证 网格一定大于或小于这个值，这个值只能指示粒度衡量，不能作为精确值衡量
        if (polygon.getArea() > meshSize * 4) {
            List<Geometry> geometryList = split(polygon);
            return geometryList.stream()
                .map(e -> split((Polygon) e, meshSize))
                .flatMap(List::stream)
                .collect(Collectors.toList());
        }
        else {
            ret.add(polygon);
        }
        return ret;
    }

    // public static Double getArea(Polygon polygon) {
    // return polygon.getArea();
    // }

    /**
     * 获得中心点坐标
     * @param wellKnownText
     * @return
     */
    public static Coordinate getCenterCoordinate(String wellKnownText) {
        Geometry geometry = wellKnownTextToGeometry(wellKnownText);
        if (geometry instanceof Polygon) {
            Polygon polygon = (Polygon) geometry;
            Coordinate centerCoordinate = polygon.getCentroid().getCoordinate();
            return centerCoordinate;
        }
        if (geometry instanceof GeometryCollection) {
            // MultiPolygon multiPolygon = (MultiPolygon) geometry;
            // // 面心
            // Centroid centroidCalculator = new Centroid(multiPolygon);
            // Coordinate centerCoordinate = centroidCalculator.getCentroid();
            // return centerCoordinate;

            GeometryCollection collection = (GeometryCollection) geometry;
            // 只获取第一个Polygon的质心
            if (!collection.isEmpty() && collection.getGeometryN(0) instanceof Polygon) {
                Polygon firstPolygon = (Polygon) collection.getGeometryN(0);
                Coordinate centerCoordinate = firstPolygon.getCentroid().getCoordinate();
                return centerCoordinate;
            }

        }
        return null;
    }

    /**
     * 计算点到多边形的最短距离的方法
     */
    public static double pointPolygonDistance(double longitude, double latitude, String wellKnownText) {
        Polygon polygon = wellKnownTextToPolygon(wellKnownText);
        Point point = toPoint(longitude, latitude);
        return point.distance(polygon) * 111 * 1000;
    }

    /**
     * 计算两个点之间的距离（单位：米）
     * @param lon1
     * @param lat1
     * @param lon2
     * @param lat2
     * @return
     */
    public static double calculateDistance(double lon1, double lat1, double lon2, double lat2) {
        final int R = 6371; // 地球半径，单位：千米

        // 将经纬度转换为弧度
        double lon1Rad = Math.toRadians(lon1);
        double lat1Rad = Math.toRadians(lat1);
        double lon2Rad = Math.toRadians(lon2);
        double lat2Rad = Math.toRadians(lat2);

        // 使用Haversine公式计算距离
        double dlon = lon2Rad - lon1Rad;
        double dlat = lat2Rad - lat1Rad;
        double a = Math.pow(Math.sin(dlat / 2), 2)
                + Math.cos(lat1Rad) * Math.cos(lat2Rad) * Math.pow(Math.sin(dlon / 2), 2);
        double c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
        double distance = R * c * 1000; // 转换为米

        return distance;
    }

    /**
     * 计算面积
     * @param geometry
     * @return
     */
    public static double getArea(Geometry geometry) {
        if (Objects.nonNull(geometry)) {
            if (geometry instanceof Polygon) {
                return getArea(geometry.getCoordinates());
            }
            if (geometry instanceof GeometryCollection) {
                GeometryCollection geometryCollection = (GeometryCollection) geometry;
                double sum = 0;
                for (int i = 0; i < geometryCollection.getNumGeometries(); i++) {
                    Geometry e = geometryCollection.getGeometryN(i);
                    if (e instanceof Polygon) {
                        sum += getArea(e.getCoordinates());
                    }
                }
                return sum;
            }
        }
        return 0;
    }

    /**
     * 高德 经纬度算面积 公式
     * @param coordinates
     * @return
     */
    private static double getArea(Coordinate[] coordinates) {
        // 将百度坐标转为高德坐标进行后续计算
        if (coordinates != null && coordinates.length > 0) {
            for (Coordinate coordinate : coordinates) {
                CoordinateUtil.Coordinate c = CoordinateUtil.bd09ToGcj02(coordinate.x, coordinate.y);
                coordinate.setX(c.getLng());
                coordinate.setY(c.getLat());
            }
        }

        // wgs84 椭球体长轴半径
        double sJ = 6378137.0;
        // 1度 (deg) = 0.017453292519943295 弧度 (rad)
        double Hq = 0.017453292519943295;
        double c = sJ * Hq;
        double d = 0;
        if (2 < coordinates.length) {
            for (int i = 0; i < coordinates.length - 1; i++) {
                Coordinate h = coordinates[i];
                Coordinate k = coordinates[i + 1];
                double u = h.x * c * Math.cos(h.y * Hq);
                double hhh = h.y * c;
                double v = k.x * c * Math.cos(k.y * Hq);
                d = d + (u * k.y * c - v * hhh);
            }
            Coordinate g1 = coordinates[coordinates.length - 1];
            Coordinate point = coordinates[0];
            double eee = g1.x * c * Math.cos(g1.y * Hq);
            double g2 = g1.y * c;
            double k = point.x * c * Math.cos(point.y * Hq);
            d += eee * point.y * c - k * g2;
        }
        // 平方千米
        return new BigDecimal(0.5 * Math.abs(d)).divide(new BigDecimal(1000000)).doubleValue();
    }

    // public static void main(String[] args) {
    // System.out.println(
    // wellKnownTextToWebStr("POLYGON((104.262044018766 31.0427551335165, 104.556985
    // 30.990947, 104.593913240945 30.6585641871206, 104.262044018766 31.0427551335165))")
    // );
    // System.out.println(
    // wellKnownTextToWebStr("MULTIPOLYGON(((104.609472697989 30.5185170046205, 104.642072
    // 30.225098, 103.618722 30.26104, 103.328965 31.206656, 103.991743095636
    // 31.0902350946145, 103.910716993188 30.6871821747449, 104.609472697989
    // 30.5185170046205)), ((104.262044018766 31.0427551335165, 104.556985 30.990947,
    // 104.593913240945 30.6585641871206, 104.262044018766 31.0427551335165)))")
    // );
    // System.out.println(
    // webStrToWellKnownText("104.262044018766,31.0427551335165;104.556985,30.990947;104.593913240945,30.6585641871206;104.262044018766,31.0427551335165")
    // );
    // System.out.println(
    // webStrToWellKnownText("104.609472697989,30.5185170046205;104.609472697989,30.5185170046205;104.609472697989,30.5185170046205;104.609472697989,30.5185170046205;104.609472697989,30.5185170046205;104.609472697989,30.5185170046205;104.609472697989,30.5185170046205;;104.556985,30.990947;104.556985,30.990947;104.556985,30.990947;104.556985,30.990947")
    // );
    // }

    /**
     * 合并图形
     * @param geometryList
     * @return
     */
    public static Geometry mergePolygons(List<Geometry> geometryList) {
        try {
            if (CollectionUtils.isEmpty(geometryList)) {
                return null;
            }
            // 0.1毫米 snapTolerance 可拖放对象离目标对象的距离低于此像素值时开始靠拢
            double snapTolerance = 0.00001;
            // 对相邻几何体进行对齐（snapping）取消用线分割占的缝
            for (int i = 0; i < geometryList.size(); i++) {
                for (int j = i + 1; j < geometryList.size(); j++) {
                    GeometrySnapper snapper = new GeometrySnapper(geometryList.get(i));
                    geometryList.set(i, snapper.snapTo(geometryList.get(j), snapTolerance));

                    snapper = new GeometrySnapper(geometryList.get(j));
                    geometryList.set(j, snapper.snapTo(geometryList.get(i), snapTolerance));
                }
            }

            // 将多边形转换为一个几何集合
            Geometry[] geometryArray = geometryList.toArray(new Geometry[0]);
            GeometryCollection geometryCollection = geometryFactory.createGeometryCollection(geometryArray);
            // 使用Union方法合并所有多边形
            Geometry mergedPolygon = geometryCollection.union();
            return mergedPolygon;
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IllegalArgumentException("合并失败！");
        }

    }

    public static Polygon ConvertLatitudeAndLongitudeFormat(String coordinatesStr) {
        // 解析经纬度数据为 Coordinate 对象数组
        String[] coordinatesArray = coordinatesStr.split(";");
        Coordinate[] coordinates = new Coordinate[coordinatesArray.length];
        for (int i = 0; i < coordinatesArray.length; i++) {
            String[] latLng = coordinatesArray[i].split(",");
            coordinates[i] = new Coordinate(Double.parseDouble(latLng[0]), Double.parseDouble(latLng[1]));
        }

        // 创建 GeometryFactory 对象
        GeometryFactory geometryFactory = new GeometryFactory();

        // 创建 LinearRing 对象
        LinearRing linearRing = geometryFactory.createLinearRing(coordinates);

        // 创建 Polygon 对象
        Polygon polygon = geometryFactory.createPolygon(linearRing);
        return polygon;
    }

	public static Coordinate getStrPointToCenterPoint(String str) {
		String[] arr = str.split(";");
		int total = arr.length;
		double X = 0, Y = 0, Z = 0;
		for (int i = 0; i < arr.length; i++) {
			double lat, lon, x, y, z;
			try {
				lon = Double.parseDouble(arr[i].split(",")[0]) * Math.PI / 180;
				lat = Double.parseDouble(arr[i].split(",")[1]) * Math.PI / 180;
			} catch (NumberFormatException e) {
				continue;
			}
			x = Math.cos(lat) * Math.cos(lon);
			y = Math.cos(lat) * Math.sin(lon);
			z = Math.sin(lat);
			X += x;
			Y += y;
			Z += z;
		}

		X = X / total;
		Y = Y / total;
		Z = Z / total;
		double Lon = Math.atan2(Y, X);
		double Hyp = Math.sqrt(X * X + Y * Y);
		double Lat = Math.atan2(Z, Hyp);

		return new Coordinate(Lon * 180 / Math.PI, Lat * 180 / Math.PI);
	}

	public static Coordinate getLineCenter(String line){
		WKTReader reader = new WKTReader();
		try {
			// 解析LINESTRING
			Geometry geometry = reader.read(line);

			// 计算中心点
			Geometry centroid = geometry.getCentroid();

			// 获取中心点的坐标
			return centroid.getCoordinate();
		}catch (ParseException e) {
			e.printStackTrace();
		}
		return null;
	}
    // public static void main(String[] args) {
    // String aa =
    // "104.02229459596911,30.619285473641956;103.9755766351166,30.61282980844764;103.97370229288,30.66631622102867;103.96963750722634,30.68669219150698;103.96433058169704,30.713294765294606;104.09579622856991,30.744319162692523;104.09584692863132,30.74433112733053;104.12138992771733,30.683959619830283;104.12237637205784,30.681628134400558;104.14176831342786,30.635794805695046;104.11149578545172,30.5361139072336;104.02229459596911,30.619285473641956";
    // System.out.println(ConvertLatitudeAndLongitudeFormat(aa).toText());
    // }

}
