package com.hncy.sjdz.base.util.cdEncryption;

import cn.hutool.http.HttpRequest;
import cn.hutool.json.JSONObject;
import cn.hutool.json.JSONUtil;
import com.hncy4cloud.hncy.common.core.exception.business.BusinessException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.util.Base64;
import java.util.Optional;

/**
 * CdEncryptionUtil
 *
 * @author fate
 * @version 0.0.1
 * @since 2024-11-19 21:38
 **/

@Slf4j
@Component
@ConditionalOnBean(ChenDEncryptionProperties.class)
public class CdEncryptionUtil {

	/**
	 * 5.1获取token
	 */
	public static final String GET_TOKEN_URI = "/ccsp/auth/app/v1/token";

	/**
	 * 5.2对称密钥（SM4）加密（数据存储机密性保护）
	 */
	public static final String ENCRYPT_SYMMETRIC_URI = "/pki/api/v6/encrypt/internal/symmetric";

	/**
	 * 5.3对称密钥（SM4）解密（数据存储机密性保护）
	 */
	public static final String DECRYPT_SYMMETRIC_URI = "/pki/api/v6/decrypt/internal/symmetric";

	/**
	 * 5.4HMAC完整性（数据存储完整性保护）
	 */
	public static final String HMAC_SYMMETRIC_URI = "/pki/api/v6/hmac/internal/symmetric";

	@Resource
	private ChenDEncryptionProperties chenDEncryptionProperties;

	public String getToken() {
		ChenDEncryptionProperties properties = getProperties();
		String serverAddr = properties.getServerAddr();
		JSONObject object = new JSONObject();
		object.set("username", properties.getUsername());
		object.set("password", properties.getPassword());

		String body = HttpRequest.post(serverAddr + GET_TOKEN_URI)
				.body(object.toString())
				.timeout(20000)
				.execute()
				.body();

		JSONObject obj = JSONUtil.parseObj(body);
		log.debug("GET_TOKEN_URI:{}", obj);
		Boolean success = obj.getBool("success");
		if (!success) {
			throw new BusinessException(obj.getStr("message"));
		}

		return obj.getJSONObject("data").getStr("accessToken");
	}

	public String encryptSM4(String token, String inData) {
		log.debug("encryptSM4-params:{},{}", token, inData);
		String inDataStr = Base64.getEncoder().encodeToString(inData.getBytes());
		HttpRequest request = buildSM4Request(ENCRYPT_SYMMETRIC_URI, inDataStr, token);
		String body = request.execute().body();
		log.debug("encryptSM4-body:{}", body);
		return getOutData(body);
	}

	public String decryptSM4(String token, String inData) {
		log.debug("decryptSM4-params:{},{}", token, inData);
		HttpRequest request = buildSM4Request(DECRYPT_SYMMETRIC_URI, inData, token);

		String body = request.execute().body();
		log.debug("decryptSM4-body:{}", body);
		String outDataBase64Str = getOutData(body);
		byte[] decodedBytes = Base64.getDecoder().decode(outDataBase64Str);
		String outDta = new String(decodedBytes);
		log.debug("outData:{}", outDta);
		return outDta;
	}

	public String hmac(String token, String inData) {
		log.debug("hmac-params:{},{}", token, inData);
		ChenDEncryptionProperties properties = getProperties();
		String serverAddr = properties.getServerAddr();

		JSONObject object = new JSONObject();
		object.set("keyName", properties.getHmacKeyName());
		object.set("algType", properties.getHmacAlgType());
		object.set("inData", inData);

		String body = HttpRequest.post(serverAddr + HMAC_SYMMETRIC_URI)
				.body(object.toString())
				.header("X-SW-Authorization-Token", token)
				.header("X-SW-Authorization-TenantCode", properties.getTenantCode())
				.header("X-SW-Authorization-AppCode", properties.getAppCode())
				.timeout(20000)
				.execute()
				.body();
		log.debug("hmac-body:{}", body);
		String outData = getOutData(body);
		log.debug("hmac-outData:{}", outData);
		return outData;
	}

	private String getOutData(String body) {
		JSONObject obj = JSONUtil.parseObj(body);

		String status = obj.getStr("status");
		if (!"200".equals(status)) {
			throw new BusinessException(obj.getStr("message"));
		}

		return obj.getJSONObject("result").getStr("outData");
	}

	public ChenDEncryptionProperties getProperties() {
		return chenDEncryptionProperties;
	}

	private HttpRequest buildSM4Request(String uri, String inData, String token) {
		JSONObject object = new JSONObject();

		ChenDEncryptionProperties properties = getProperties();

		String serverAddr = properties.getServerAddr();


		object.set("keyName", properties.getSymmetricKeyName());
		object.set("algType", properties.getSymmetricAlgType());
		object.set("paddingType", properties.getSymmetricPaddingType());
		object.set("iv", properties.getSymmetricIv());
		object.set("inData", inData);
		return HttpRequest.post(serverAddr + uri)
				.body(object.toString())
				.header("X-SW-Authorization-Token", token)
				.header("X-SW-Authorization-TenantCode", properties.getTenantCode())
				.header("X-SW-Authorization-AppCode", properties.getAppCode())
				.timeout(20000);
	}

	public boolean enable() {

		return Optional.ofNullable(getProperties().getEnable()).orElse(false);
	}

}
